/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package de.mfo.jsurfer.gui;

import javax.swing.*;
import javax.swing.event.*;
import java.awt.*;
import java.awt.event.*;

import de.mfo.jsurfer.rendering.*;
import de.mfo.jsurfer.rendering.cpu.*;
import de.mfo.jsurfer.debug.*;
import java.awt.image.*;
import javax.vecmath.*;

import de.mfo.jsurfer.algebra.*;
import de.mfo.jsurfer.parser.*;

/**
 *
 * @author Christian Stussak <christian at knorf.de>
 */
public class TestApplet extends JApplet
{

    CPUAlgebraicSurfaceRenderer c;
    int[] colorBuffer;
    MemoryImageSource mis;
    int width;
    int height;
    MausZeuch mz;
    String expression;
    float scale;
    
    class MausZeuch extends MouseInputAdapter
    {
        Point oldMousePos;
                
        public void mousePressed( MouseEvent me )
        {
            oldMousePos = me.getPoint();
        }
        
        public void mouseDragged( MouseEvent me )
        {
            Globals.mousePos = me.getPoint();
            Globals.pause = me.isShiftDown();
            
            draw( oldMousePos, me.getPoint() );
        }
        
        public void draw( Point oldMousePos, Point newMousePos )
        {
            long startTime = System.currentTimeMillis();
            Matrix4f rotY = new Matrix4f();
            rotY.setIdentity();
            rotY.rotY( ( float ) ( Math.PI * ( newMousePos.x - oldMousePos.x ) ) / 180.0f );
            
            Matrix4f rot = new Matrix4f();
            rot.rotX( ( float ) -( Math.PI * ( newMousePos.y - oldMousePos.y ) ) / 180.0f );
            rot.mul( rotY );
            
            Matrix4f scale = new Matrix4f();
            scale.setIdentity();
            scale.setScale( 1.0f / TestApplet.this.scale );
                                    
            Matrix4f translation = new Matrix4f();
            translation.setIdentity();
            translation.setTranslation( new Vector3f( 0f, 0.0f, 0.0f ) );
            
            Matrix4f surfaceTransform = rot;
            surfaceTransform.mul( scale );
            
            c.setSurfaceTransform( surfaceTransform );
            c.setTransform( translation );
            c.draw( colorBuffer, width, width );
            mis.newPixels();
            //f.repaint();
            
            paint( getGraphics() );
        }                
    }
    
    public void createAndShowGUI()
    {        
        width = 200;
        height = 200;
        
        
        mz = new MausZeuch();
        addMouseListener( mz );
        addMouseMotionListener( mz );
        
        // create buffer for image data
        colorBuffer = new int[ width * height ];
        
        // create Java-MemoryImageSource
        mis = new MemoryImageSource( width, height, colorBuffer, 0, width );
        mis.setAnimated(true);
        mis.setFullBufferUpdates(true);
        
        // create corresponding image for display
        Image screenImage = Toolkit.getDefaultToolkit().createImage( mis );
        
        // put image into a JLabel
        JLabel l = new JLabel( new ImageIcon( screenImage ) );
        getContentPane().add( l );
        
        setVisible(true);
    }
    
    public void init()
    {
        readParameters();
        createAndShowGUI();
        c = new CPUAlgebraicSurfaceRenderer();
        c.setBackgroundColor( new Color3f( 1.0f, 1.0f, 1.0f ) );
        
        LightSource lightSource0 = new LightSource();
        lightSource0.setPosition( new Point3f( -100f, 100f, 100f ) );
        lightSource0.setIntensity( 0.5f );
        lightSource0.setColor( new Color3f( 1f, 1f, 1f ) );
        
        LightSource lightSource1 = new LightSource();
        lightSource1.setPosition( new Point3f( 113.1187f, 148.2476f, 183.2476f ) );
        lightSource1.setIntensity( 0.7f );
        lightSource1.setColor( new Color3f( 1f, 1f, 1f ) );

        LightSource lightSource2 = new LightSource();
        lightSource2.setPosition( new Point3f( 0f, -100f, 100f ) );
        lightSource2.setIntensity( 0.3f );
        lightSource2.setColor( new Color3f( 1f, 1f, 1f ) );
        
        c.setLightSource( 0, lightSource0 );
        c.setLightSource( 1, lightSource1 );
        c.setLightSource( 2, lightSource2 );
        
        Material frontMaterial = new Material();
        frontMaterial.setColor( new Color3f( 1.0f, 0.8f, 0.4f ) );
        frontMaterial.setAmbientIntensity( 0.4f );
        frontMaterial.setDiffuseIntensity( 0.8f );
        frontMaterial.setSpecularIntensity( 0.8f );
        frontMaterial.setShininess( 50 );

        Material backMaterial = new Material();
        backMaterial.setColor( new Color3f( 0.8f, 0.4f, 0.0f ) );
        backMaterial.setAmbientIntensity( 0.4f );
        backMaterial.setDiffuseIntensity( 0.8f );
        backMaterial.setSpecularIntensity( 0.8f );
        backMaterial.setShininess( 50 );
        
        c.setFrontMaterial( frontMaterial );
        c.setBackMaterial( backMaterial );
        
        Camera camera = new Camera();
        camera.lookAt( new Point3f( 0f, 0f, 1f ), new Point3f( 0f, 0f, 0f ), new Vector3f( 0f, 1f, 0f ) );
        
        c.setCamera( camera );
                
        try
        {
            c.setSurfaceExpression( AlgebraicExpressionParser.parse( this.expression ) );
        }
        catch( Exception e )
        {
            e.printStackTrace();
        }
        
        mz.draw( new Point(), new Point() );
    }
    
    public String[][] getParameterInfo()
    {
        String pinfo[][] = {
                 {"expression",    "Sting",    "formula of this algebraic surface"},
                 {"scale", "float", "scale of the surface"}
         };
        return pinfo;
    }
    
    private void readParameters()
    {
        this.expression = getParameter( "expression" );
        if( this.expression == null )
            this.expression = "x";
        String scale = getParameter( "scale" );
        try
        {
            this.scale = Float.parseFloat( scale );
        }
        catch( NumberFormatException nfe )
        {
            this.scale = 1.0f;
        }
    }
}
